// Copyright (c) 2014, The Shelf Route project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf_route;

import 'package:shelf/shelf.dart';
import 'package:path/path.dart' as p;
import 'package:matcher/matcher.dart';
import 'src/preconditions.dart';
import 'package:uri/uri.dart';

/// A [Handler] that routes to other handlers based on the incoming request
/// (path and method). Note there can be several layers of routers to facilitate
/// modularity.
///
/// When the Router routes a request to a handler it adjusts the requests as follows:
/// * the routes path is removed from the start of the requests pathInfo
/// * the routes path is added to the end of the requests scriptName
///
/// e.g if original request had path info of `/banking' and scriptName was
/// `/abc` then when routing a request of `/banking/accounts` the new request
/// past to the handler will have a pathInfo of `/accounts` and scriptName of
/// `/abc/banking`
///
///
Router router() => new Router();

const String _PATH_VARIABLE_PARAM_PREFIX = 'shelf.route.';

Map<String, Object> getPathVariables(Request request) {
  // TODO(anders): replace with request.extraParams when supported in shelf
  final params = request.headers;
  final keys = params.keys.where(
      (k) => k.startsWith(_PATH_VARIABLE_PARAM_PREFIX));

  final pathVars = <String, Object>{};
  keys.forEach((k) {
    pathVars[k.substring(_PATH_VARIABLE_PARAM_PREFIX.length)] = params[k];
  });

  return pathVars;
}

class Router {
  final List<_Route> _routes = <_Route>[];
  final _Route _fallbackRoute;

  Router._internal(this._fallbackRoute);

  Router([Handler fallbackHandler])
      : this._fallbackRoute = new _Route(fallbackHandler != null ?
          fallbackHandler : _send404, null, null);

  void addRoute(Handler handler, {String path, String method}) {
    _routes.add(new _Route(handler, path, method));
  }

  void get(String path, Handler handler) {

    addRoute(handler, path: path, method: 'GET');
  }

  void post(String path, Handler handler) {

    addRoute(handler, path: path, method: 'POST');
  }

  Router addChildRouter(String path) {
    final child = new Router._internal(_fallbackRoute);
    addRoute(child.handler, path: path);
    return child;
  }

  Handler get handler => _handleRequest;

  dynamic _handleRequest(Request request) {
    final route = _routes.firstWhere((r) => r.canHandle(request),
        orElse: () => _fallbackRoute);
    return route.handle(request) ;
  }
}


class _Route {
  final Handler handler;
  final UriPattern _pathPattern;
  final String _method;

  _Route._internal(this.handler, this._pathPattern, this._method) {
    checkThat(handler, isNotNull, 'handler cannot be null');
    checkThat(_pathPattern, isNotNull, '_pathPattern cannot be null');
  }

  factory _Route(Handler handler, String path, String method) {

    // Note: we change '/' to '' as UriParser strips trailing '/'
    // and a '' path won't match agains a pattern of '/'
    final nonNullPath = path == null ? '' : path == '/' ? '' : path;
    final pathPattern = new UriParser(new UriTemplate(nonNullPath));

    return new _Route._internal(handler, pathPattern, method);

  }

  bool canHandle(Request request) {
    return
        (_method == null || request.method == _method)
        &&
        _pathPattern.matches(request.url);
  }

  dynamic handle(Request request) {
    // create a new request with the pathInfo and scriptName adjusted relative
    // to route path
    final newRequest = _createNewRequest(request);

    return handler(newRequest);
  }

  Request _createNewRequest(Request request) {
    final uriMatch = _pathPattern.match(request.url);
//    uriMatch.

    final requestSegs = request.url.pathSegments;

    final skipCount = request.scriptName.isNotEmpty && requestSegs.isNotEmpty
        && requestSegs.first == '/' ? 1 : 0;
    final extraScriptPath = requestSegs.skip(skipCount).take(
        requestSegs.length - uriMatch.rest.pathSegments.length);
    var newScriptName = _adjustPath(p.joinAll(p.split(request.scriptName)
        ..addAll(extraScriptPath)));

    var newPathInfo = _adjustPath(uriMatch.rest.path);

    if (newPathInfo == '' && newScriptName == '') {
      newPathInfo = '/';
    }

    final newUrl = Uri.parse(newPathInfo);

    var newHeaders = request.headers;
    if (uriMatch.parameters.isNotEmpty) {
      newHeaders = {}..addAll(request.headers);

      uriMatch.parameters.forEach((k, v) {
          newHeaders['$_PATH_VARIABLE_PARAM_PREFIX${k}'] = v;
      });
    }

//    final newRequest = request.copyWith(pathInfo: newPathInfo,
//        scriptName: newScriptName);
    return new Request(request.method, request.requestedUri,
        protocolVersion: request.protocolVersion, headers: newHeaders,
        url: newUrl, scriptName: newScriptName,
        body: request.read());

  }

  String _adjustPath(String path) {
    if (path == '/') {
      return '';
    }
    if (path.isNotEmpty && !path.startsWith('/')) {
      return p.join('/', path);
    }

    return path;
  }

}

Response _send404(Request req) {
  return new Response.notFound("Not Found");
}


